import argparse

from sate import VERSION, parse


def find_satefile():
    return '.satefile'


def load_satefile(path):
    with open(path) as rfile:
        return parse.load_satefile(rfile)


def print_targets(satefile):
    for target in satefile.targets:
        print(target)


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-l', '--list', help='list targets', action='store_true')
    parser.add_argument('--version', action='store_true')
    parser.add_argument('target', nargs='?')
    parser.add_argument('args', nargs=argparse.REMAINDER)
    args = parser.parse_args()

    satefile = load_satefile(find_satefile())

    if args.version:
        print(VERSION)
    elif args.list:
        print_targets(satefile)
    else:
        satefile.run(args.target, args.args)



from parsy import regex, seq, string

from sate import types

# pylint: disable=invalid-name

OpenParen = string('(')
CloseParen = string(')')

OpenBrack = string('[')
CloseBrack = string(']')

Whitespace = regex(r'[ \t]+')
Ident = regex(r'[^ \t\(\)\[\]]+')

ArgList = OpenParen >> Ident.sep_by(Whitespace) << CloseParen
Call = seq(Ident, ArgList).combine(types.Call) | Ident.map(types.Call)

CallList = Call.sep_by(Whitespace)

CommandTag = OpenBrack >> CallList << CloseBrack


def _make_target(name, _, directives):
    """Make a Target from name and directives."""
    return types.Target(name=name, directives=directives)


TargetTagContentWithCalls = (seq(Ident, Whitespace, CallList)
                             .combine(_make_target))
TargetTagContentSimple = Ident.map(types.Target)

TargetTagContent = TargetTagContentWithCalls | TargetTagContentSimple
TargetTag = OpenBrack >> TargetTagContent << CloseBrack



import collections
import copy
import subprocess

import attr
import toposort

DEPS_KEY = 'deps'


@attr.s(frozen=True, slots=True)
class Call(object):
    name = attr.ib(validator=attr.validators.instance_of(str))
    args = attr.ib(default=attr.Factory(list))


@attr.s(frozen=True, slots=True)
class Command(object):
    text = attr.ib()
    directives = attr.ib(default=attr.Factory(list))

    def with_directives(self, directives):
        return Command(text=self.text, directives=directives)

    def run(self, extra_args):
        cmd = ' '.join([self.text] + extra_args)
        print(cmd)
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError as err:
            print('failed with exit code {}'.format(err.returncode))


@attr.s(frozen=True, slots=True)
class Comment(object):
    text = attr.ib()


def separate_deps_from_directives(original_directives):
    directives = []
    deps = []
    for directive in original_directives:
        if directive.name == DEPS_KEY:
            deps += directive.args
        else:
            directives.append(directive)
    return deps, directives


@attr.s(frozen=True, slots=True)
class Target(object):
    name = attr.ib()
    commands = attr.ib(default=attr.Factory(list))
    directives = attr.ib(default=attr.Factory(list))

    def add_command(self, command):
        return Target(name=self.name, commands=self.commands + [command])

    def deps(self):
        for directive in self.directives:
            if directive.name == DEPS_KEY:
                return directive.args
        return []

    def run(self, args):
        for command in self.commands:
            command.run(args)

    def with_deps(self, *new_deps):
        deps, directives = separate_deps_from_directives(self.directives)
        commands = copy.copy(self.commands)
        for dep in new_deps:
            if dep not in deps:
                deps.append(dep)
        directives.append(Call(DEPS_KEY, deps))
        return Target(name=self.name, commands=commands, directives=directives)


def target_list_to_dict(lst):
    lst = lst or []
    dct = collections.OrderedDict()
    for target in lst:
        if target.name in dct:
            raise KeyError('duplicate target name: ' + target.name)
        dct[target.name] = target
    return dct


def make_target_graph(all_targets, original_target):
    graph = {}

    def init_node(target):
        if target not in graph:
            graph[target] = set()

    stack = [original_target]
    while stack:
        target = stack.pop()
        init_node(target)
        deps = all_targets[target].deps()
        for dep in deps:
            init_node(dep)
            graph[target].add(dep)
        stack += deps

    return graph


@attr.s(frozen=True, slots=True)
class Satefile(object):
    targets = attr.ib(
        default=attr.Factory(collections.OrderedDict),
        validator=attr.validators.instance_of(collections.abc.Mapping),
        convert=target_list_to_dict)

    def run(self, target_name, args):
        if target_name not in self.targets:
            exit('error: unknown target')
        for dep in reversed(self.run_order(target_name)):
            self.targets[dep].run(args)

    def run_order(self, target):
        graph = make_target_graph(self.targets, target)
        return list(toposort.toposort_flatten(graph))



import logging

from sate import rules, types

LOG = logging.getLogger(__name__)


class ParseError(ValueError):
    pass


def parse_line(line):
    comment_start = line.find('#')
    LOG.debug('comment_start: %s', comment_start)
    has_comment = comment_start != -1
    LOG.debug('has_comment: %s', has_comment)

    pre_comment_end = comment_start if has_comment else len(line)
    LOG.debug('pre_comment_end: %s', pre_comment_end)
    tag_start = line.find('[', 0, pre_comment_end)
    tag_end = line.find(']', 0, pre_comment_end)
    has_tag = tag_start != -1

    tag_text = None

    # Get tag
    if has_tag and tag_end == -1:
        raise ParseError('missing "]"')
    elif (not has_tag) and tag_end != -1:
        raise ParseError('unexpected "]"')
    elif tag_end < tag_start:
        raise ParseError('unexpected "]"')
    elif has_tag:
        tag_text = line[tag_start:tag_end + 1]

    # Get command text
    post_tag_start = 0 if not has_tag else tag_end + 1
    command_text = line[post_tag_start:pre_comment_end].strip()
    command = types.Command(text=command_text)

    if tag_text and command.text:
        yield command.with_directives(list(rules.CommandTag.parse(tag_text)))
    elif tag_text:
        yield rules.TargetTag.parse(tag_text)
    elif command.text:
        yield command

    # Get comment
    if has_comment:
        yield types.Comment(line[comment_start + 1:])


def compose(elements):
    target = None
    for elem in elements:
        LOG.debug('target: %s, elem: %s', target, elem)
        if isinstance(elem, types.Comment):
            yield elem
        elif isinstance(elem, types.Target):
            if target is None:
                target = elem
            else:
                yield target
                target = elem
        elif isinstance(elem, types.Command):
            if target is None:
                raise ParseError('invalid command: outside of target')
            else:
                target = target.add_command(elem)
        else:
            raise TypeError('unexpected element type: {}'.format(elem))
    if target:
        yield target


def parse_file(rfile):
    def gen():
        for line in rfile.readlines():
            parsed = parse_line(line)
            for elem in parsed:
                yield elem

    return compose(gen())


def load_satefile(rfile):
    targets = []
    for elem in parse_file(rfile):
        if isinstance(elem, types.Target):
            targets.append(elem)
    return types.Satefile(targets)
